classdef simpleModelENPSVM < simpleModel
    % SIMPLEMODELENPSVM This class inherits from simpleModel. It contains
    % particular methods to solve binary-C-SVM for non positive definite kernels, using eigvalues.
    % This class support C-SVM solver, warm-start, 2D solution plot, and
    % cross-validation error-rate evaluation (see simpleModel.crossValError).
    % See also : simpleModel, simpleData
    
    % Create by Galle Loosli
    % 16/02/2011
    % Last Modified 16/02/2011 by G.L.
    
    properties
       negs % indices of negative eigenvalues in the decomposition
       V % eigenvectors needed to get the solution in the initial kernel space
    end
    
    methods
        % contructor
        function SM = simpleModelENPSVM(SD,C,nbEig,Il,Iu,Iw)
            % simpleModelCSVM : constructor.
            % Input :   SD : a simpleData object
            %           C : Slack value (can be of length 1 for all points
            %           equally penalized, or one value by class, or one
            %           value per training point)
            %           nbEig : nb of eigenvalues computed for
            %           decomposition
            %           Il : indices list of non support vectors for warm-start
            %           Iu : indices list of upper bounded support vectors for
            %           warm-start
            %           Iw : indices list of non-bounded support vectors
            % Number of input arguments must be at least 2 (simpleData
            % object and simpleProblem object) or exactly 5 (for warm-start,
            % one can provide the initial repartition of support vectors)
            % Output :   SM : a simpleModelCSVM object
            
            assert(nargin==3 || nargin==6,'simpleModelENPSVM:wrongInput','Requires 3 or 6 input arguments');
            if nargin<3
                error('not enough input');
            else
                
                SD.getKernelValues(1:SD.n); % pre-compute kernel
                [SM.V D] = eigs(SD.kernelCache(1:SD.n,1:SD.n),min(nbEig,SD.n-2));
                SM.negs = find(diag(D)<0);
                SD.kernelCache(1:SD.n,1:SD.n) = SM.V*abs(D)*SM.V';
                SM.SD = SD;
                SM.C = C;
                SM.Slacks = zeros(SD.n,1);
                if nargin==5
                    SM.Il = Il;
                    SM.Iu = Iu;
                    SM.Iw = Iw;
                    SM.oldalpha = zeros(length(Iw),1);
                else
                    SM.Iw = [find(SD.trainlab==-1,1),find(SD.trainlab==1,1)];
                    SM.Iu = [];
                    SM.Il = 1:SD.n;
                    SM.Il = setdiff(SM.Il,SM.Iw);
                    SM.oldalpha = zeros(2,1);
                end
                SM.labs = sort(unique(SD.trainlab));
                if length(C)==length(SD.trainlab)
                    SM.Slacks = C;
                elseif length(C)==length(SM.labs)
                    for i=1:length(SM.labs)
                        SM.Slacks(SD.trainlab==SM.labs(i)) = C(i);
                    end
                else
                    SM.Slacks = C(1)*ones(size(SD.trainlab));
                    
                end
                SM.activated = 0;
                SM.converged = 0;
            end
        end
        
        
        function set.negs(obj,val)
            obj.negs = val;
        end
        
        function val = get.negs(obj)
            val = obj.negs;
        end
        
        function set.V(obj,val)
            obj.V = val;
        end
        
        function val = get.V(obj)
            val = obj.V;
        end
        
        
        function solveLinearSystem(obj)
            ce = ones(1,length(obj.Iw));
            be = 0;
            indd = obj.SD.getKernelValues(obj.Iw,obj.Iu);
            if ~isempty(obj.Iu)
                ce = ce - (obj.Slacks(obj.Iu))'*obj.SD.kernelCache(obj.Iu,indd);
                be = be - obj.Slacks(obj.Iu)'*obj.SD.trainlab(obj.Iu);
            end
            if obj.optimType==1
                G = chol(obj.SD.kernelCache(obj.Iw,indd)+obj.kktgap*eye(length(obj.Iw)));
                M = obj.SD.trainlab(obj.Iw)'*(G\(G'\obj.SD.trainlab(obj.Iw)));
                obj.b = M\(obj.SD.trainlab(obj.Iw)'*(G\(G'\ce'))-be);
                obj.alpha = G\(G'\(ce'-obj.SD.trainlab(obj.Iw)*obj.b));
            else
                error('not implemented yet');
            end
        end
        
        function print2D(obj)
            assert(size(obj.SD.trainvec,1)==2,'simpleModelENPSVM:print2D','must be 2D data');
            scale = 2;
            [xtesta1,xtesta2]=meshgrid([-scale:0.1:scale],[-scale:0.1:scale]);
            [na,nb]=size(xtesta1);
            xtest1=reshape(xtesta1,1,na*nb);
            xtest2=reshape(xtesta2,1,na*nb);
            xtest=[xtest1;xtest2];
            if obj.converged==0
                obj.SV = obj.SD.trainvec(:,[obj.Iw,obj.Iu]);
                f = kernelEval(obj,xtest);
                ypred = ([obj.alpha;obj.Slacks(obj.Iu)].*obj.SD.trainlab([obj.Iw,obj.Iu]))'*f+obj.b;
            else
                f = kernelEval(obj,xtest);
                ypred = obj.alpha'*f+obj.b;
            end
            ypredmat=reshape(ypred,na,nb);
            
            %hold off;
            plot(obj.SD.trainvec(1,obj.SD.trainlab==1),obj.SD.trainvec(2,obj.SD.trainlab==1),'r+');
            hold on;
            plot(obj.SD.trainvec(1,obj.SD.trainlab==-1),obj.SD.trainvec(2,obj.SD.trainlab==-1),'bx');
            pp = zeros(size(obj.SD.trainlab));
            pp([obj.Iw,obj.Iu]) = 1;
            pp = pp.*obj.SD.trainlab;
            plot(obj.SV(1,pp>0),obj.SV(2,pp>0),'or','MarkerSize',8,'LineWidth',2); 
            plot(obj.SV(1,pp<0),obj.SV(2,pp<0),'ob','MarkerSize',8,'LineWidth',2);   
            plot(obj.SV(1,obj.alpha.*obj.SD.trainlab>0),obj.SV(2,obj.alpha.*obj.SD.trainlab>0),'ob');
            plot(obj.SV(1,obj.alpha.*obj.SD.trainlab<0),obj.SV(2,obj.alpha.*obj.SD.trainlab<0),'or');         
            [c,h] = contour(xtesta1,xtesta2,ypredmat,[ 0 0],'g');
            set(h,'LineWidth',2);
            [c,h] = contour(xtesta1,xtesta2,ypredmat,[ -1 1],'y.');
            set(h,'LineWidth',1);
            drawnow;
            legend('class +', 'class -', 'SV+', 'SV-', 'decision function','margins');
            
        end
        
        
        
        function print2DMulti(obj,SMt)
            assert(size(obj.SD.trainvec,1)==2,'simpleModelCSVM:print2D','must be 2D data');
            assert(nargin==2);
            scale = 2;
            [xtesta1,xtesta2]=meshgrid([-scale:0.1:scale],[-scale:0.1:scale]);
            [na,nb]=size(xtesta1);
            xtest1=reshape(xtesta1,1,na*nb);
            xtest2=reshape(xtesta2,1,na*nb);
            xtest=[xtest1;xtest2];
            if length(SMt)==length(unique(obj.SD.trainlab));
                ypred = obj.test1vsAll(SMt,xtest);
            else
                ypred = obj.test1vs1(SMt,xtest);
            end
            ypredmat=reshape(ypred,na,nb);
            %labs = unique(obj.SD.trainlab);
            hold on;
            [c,h] = contour(xtesta1,xtesta2,ypredmat);
            set(h,'LineWidth',2);
            drawnow;
            
        end
        
        function finish(obj)
            
%             obj.alpha = [obj.alpha;obj.Slacks(obj.Iu)];
%             obj.alpha = obj.V([obj.Iw,obj.Iu],:)'*obj.alpha;
%             obj.alpha(obj.negs) = -obj.alpha(obj.negs);
%             obj.alpha = (obj.V*obj.alpha).*obj.SD.trainlab;
%             if not(strcmp(obj.SD.kernelType,'precomputed'))
%                 obj.SV = obj.SD.trainvec;
%                 
%             else
%                 %disp('no training data can be stored for later use :
%                 %pre-computed matrix');
%                 obj.Iw= 1:obj.SD.n;
%                 obj.Iu = [];
%             end
            f = tanh(obj.SD.kernelParam(1) * obj.SD.trainvec' *...
                obj.SD.trainvec + obj.SD.kernelParam(2));
            alpha = zeros(size(obj.SD.trainlab));
            alpha(obj.Iw) = obj.alpha;
            alpha(obj.Iu) = obj.Slacks(obj.Iu);
            obj.alpha = f([obj.Iw],:)\(obj.SD.kernelCache([obj.Iw],1:obj.SD.n)*alpha);
            obj.b = 0;
            %obj.alpha = obj.alpha.*obj.SD.trainlab;
            if not(strcmp(obj.SD.kernelType,'precomputed'))
                obj.SV = obj.SD.trainvec;
                
            else
                %disp('no training data can be stored for later use :
                %pre-computed matrix');
                obj.Iw= 1:obj.SD.n;
                obj.Iu = [];
            end
        end
        
                
        function reinit(obj)
            obj.Iw = [find(obj.SD.trainlab==-1,1),find(obj.SD.trainlab==1,1)];
            obj.Iu = [];
            obj.Il = 1:obj.SD.n;
            obj.Il = setdiff(obj.Il,obj.Iw);
            obj.oldalpha = zeros(2,1);
            
            if length(obj.Slacks)~=length(obj.SD.trainlab)
                if length(obj.C)==length(obj.SD.trainlab)
                    obj.Slacks = obj.C;
                elseif length(obj.C)==length(obj.labs)
                    for i=1:length(obj.labs)
                        obj.Slacks(obj.SD.trainlab==obj.labs(i)) = obj.C(i);
                    end
                else
                    obj.Slacks = obj.C(1)*ones(size(obj.SD.trainlab));
                    
                end
            end
            obj.activated = 0;
            obj.converged = 0;
            obj.SD.getKernelValues(1:obj.SD.n); % pre-compute kernel
            [obj.V D] = eigs(obj.SD.kernelCache(1:obj.SD.n,1:obj.SD.n),min(length(obj.V),obj.SD.n-2));
            obj.negs = find(diag(D)<0);
            obj.SD.kernelCache(1:obj.SD.n,1:obj.SD.n) = obj.V*abs(D)*obj.V';
        end
        
        function SMt = train1vs1(obj,SMt)
            nbClass = length(obj.labs);
            k = 0;
            if nargin==1
                for i=1:nbClass
                    for j=i+1:nbClass
                        k=k+1;
                        ind = sort([find(obj.SD.trainlab==obj.labs(i));find(obj.SD.trainlab==obj.labs(j))]);
                        tl = obj.SD.trainlab(ind);
                        tl(tl==obj.labs(i))=-1;
                        tl(tl==obj.labs(j))=1;
                        if strcmp(obj.SD.kernelType,'precomputed')
                            tt = obj.SD.trainvec(ind,ind);
                        else
                            tt = obj.SD.trainvec(:,ind);
                        end
                        SDt{k} = simpleData(tt,tl);
                        SDt{k}.kernelParam = obj.SD.kernelParam;
                        SDt{k}.kernelType = obj.SD.kernelType;
                        SMt{k} = feval(class(obj),SDt{k},obj.Slacks(ind),size(obj.V,2));
                        SMt{k}.train();
                        
                    end
                end
            elseif nargin==2
                assert(length(SMt)==(nbClass-1)*nbClass/2);
                for i=1:nbClass
                    for j=i+1:nbClass
                        k=k+1;
                        if (sum(SMt{k}.SD.kernelParam-obj.SD.kernelParam)==0|| not(strcmp(SMt{k}.SD.kernelType,obj.SD.kernelType)))
                            SMt{k}.SD.clear;
                            SMt{k}.SD.kernelType = obj.SD.kernelType;
                            SMt{k}.SD.kernelParam = obj.SD.kernelParam;
                        end
                        ind = sort([find(obj.SD.trainlab==obj.labs(i));find(obj.SD.trainlab==obj.labs(j))]);
                        
                        SMt{k} = feval(class(obj),SMt{k}.SD,obj.Slacks(ind),size(SMt{k}.V,2),SMt{k}.Il,SMt{k}.Iu,SMt{k}.Iw);
                        SMt{k}.train();
                    end
                end
            end
        end
        
        
        
    end
    
end
